IMPLEMENTATION MODULE Environment;
(************************************************************************)
(*                                                                      *)
(*      Version 1.00    08 September 1988       D. C. Sessions          *)
(*      Version 2.00    19 December  1988       D. C. Sessions          *)
(*                                                                      *)
(************************************************************************)

  FROM SYSTEM
    IMPORT ADR, ADDRESS;
  FROM psp
    IMPORT PSP, PSPPtr;
  FROM Strings
    IMPORT Assign, Length, Concat, Copy;
  FROM BlockOps
    IMPORT BlockMoveForward;
  IMPORT ASCII;
  IMPORT SYSTEM;

  TYPE
    EnvironmentString = ARRAY[ 0..32767 ] OF CHAR;
    LongString        = ARRAY[ 0..1023 ] OF CHAR;
    ShortString       = ARRAY[ 0..255 ] OF CHAR;
    PSPaddress        = POINTER TO PSP;
    EnvPointer        = POINTER TO EnvironmentString;

  VAR
    Depth     : INTEGER; (* # of environments in chain *)

  PROCEDURE EndOf( VAR env : ARRAY OF CHAR ) : INTEGER;
    (* Returns offset of empty string which terminates an environment *)

    VAR
      i : INTEGER;

    BEGIN (* EndOf *)
      i := 0;
      WHILE env[ i ] # ASCII.nul DO
        WHILE env[ i ] # ASCII.nul DO
          INC( i );
          END (* WHILE *);
        INC( i );
        END (* WHILE *);
      RETURN i;
      END EndOf;
(*
  NonPortableFilth: Here we do the dirty work
*)
  MODULE NonPortableFilth; (* Incredibly grungy low-level stuff *)

    FROM SYSTEM
      IMPORT ADR, ADDRESS, BYTE, DOSCALL, SWI, SETREG, WORD, AX, BX;
    IMPORT EndOf, PSPPtr, PSPaddress;

    EXPORT Resize, Dive, SegSize;

    TYPE
      (*$A-*)
      MCB = RECORD (* MS-DOS Memory Control Block: UNDOCUMENTED *)
              ChainID : BYTE;
              ProcID  : WORD;
              BlkSize : CARDINAL;
              END (* MCB *);
      (*$A=*)

    PROCEDURE SegSize( arg : ADDRESS ) : CARDINAL;
      (* Return allocated size of segment pointed to by 'arg' *)

      VAR
        MCBaddr : ADDRESS;
        MCBloc  : POINTER TO MCB;

      BEGIN (* SegSize *)
        MCBaddr.OFFSET := 0;
        MCBaddr.SEGMENT := arg.SEGMENT - 1;
        MCBloc := MCBaddr;
        RETURN 16 * MCBloc^.BlkSize;
        END SegSize;

    PROCEDURE Resize( VAR env      : ARRAY OF CHAR;
                          change   : INTEGER;
                          landlord : ADDRESS ) : BOOLEAN;
      (*   Resize increases the allocated size of the segment containing *)
      (*   'env' enough to add 'change' to its length.                   *)
      (*   To preserve the memory arena, the change is ordered           *)
      (*   under the ownership of the PSP addressed by 'landlord'.       *)
      (*   The return value is TRUE iff the attempt fails.               *)
      (*   This whole disgusting business is utterly nonportable.        *)

      PROCEDURE SetPSP( arg : ADDRESS );
        (*   SetPSP points DOS at a (possibly foreign) PSP, so that   *)
        (*   any memory allocated is 'owned' by that program.         *)

        BEGIN
          SETREG( BX, arg.SEGMENT );
          SETREG( AX, 5000H );
          SWI( 21H );
          END SetPSP;

      VAR
        ThisSeg   : ADDRESS;
        ByteSize,
        newsize,
        ErrorCode : CARDINAL;

      BEGIN (* Resize *)
        IF change <= 0
          THEN
            RETURN FALSE;                 (* Don't shrink the environment *)
          ELSE
            ThisSeg := ADR( env );
            ByteSize := CARDINAL( change ) + SegSize( ThisSeg );
            newsize := ( ByteSize DIV 16 ) + 1;
            SetPSP( landlord );                (* Pretend to be the owner *)
            DOSCALL( 4AH, ThisSeg,
                     newsize, ErrorCode );            (* Resize the block *)
            SetPSP( PSPPtr );            (* Now back to our true identity *)
            RETURN ErrorCode # 0;
          END (* IF *);
        END Resize;

  PROCEDURE Dive( arg : PSPaddress ) : PSPaddress;
    (* Return the address of the parent PSP *)

    VAR
      next : ADDRESS;

    BEGIN (* Dive *)
      next.OFFSET := 0;                       (* PSPs are segment-aligned *)
      next.SEGMENT := CARDINAL( arg^.PspLink );
      RETURN next;
      END Dive;

    BEGIN (* NonPortableFilth *)
      END NonPortableFilth;
(*
  SetLevel locates the PSP of the requested level
*)
  VAR
    LastLevel : INTEGER;    (* Previous search level *)
    LastPSP   : PSPaddress; (* Previous PSP location *)

  PROCEDURE SetLevel( WhereTo : INTEGER ) : PSPaddress;
    (* Return the address of the requested PSP *)

    VAR
      RelLevel : INTEGER;

    BEGIN (* SetLevel *)
      IF WhereTo <= 0
        THEN RelLevel := WhereTo;
        ELSE RelLevel := WhereTo - Depth;
        END (* IF WhereTo *);
      IF RelLevel > LastLevel THEN
        LastLevel := 0;
        LastPSP   := PSPaddress( PSPPtr );
        END (* IF RelLevel *);
      WHILE RelLevel < LastLevel DO
        DEC( LastLevel );
        LastPSP := Dive( LastPSP );
        END (* WHILE *);
      RETURN LastPSP;
      END SetLevel;

  PROCEDURE CapString( VAR arg : ARRAY OF CHAR );
    (* The obvious string capitalization function. *)

    VAR
      i   : CARDINAL;

    BEGIN (* CapString *)
      i := 0;
      WHILE   ( i <= HIGH( arg ) )
          AND ( arg[ i ] # ASCII.nul ) DO
        arg[ i ] := CAP( arg[ i ] );
        INC( i );
        END (* WHILE *);
      END CapString;

  PROCEDURE FindString(  key : ARRAY OF CHAR;
                        base : ADDRESS ) : ADDRESS;
    (*   FindString searches the environment at 'base'       *)
    (*   for a string beginning with '<key>='.  If no such   *)
    (*   string is found, FindString returns NIL.            *)
    (*   Otherwise, it returns a pointer to the string.      *)

    PROCEDURE Match(      key : ARRAY OF CHAR;
                     VAR lock : ARRAY OF CHAR ) : BOOLEAN;
      (* TRUE iff 'key' matches the first part of 'lock' *)

      VAR
        i       : CARDINAL;
        MaxKey  : CARDINAL;
        OkSoFar : BOOLEAN;

      BEGIN (* Match *)
        i := 0;
        MaxKey := Length( key );
        OkSoFar := ( MaxKey < Length( lock ) );
        WHILE OkSoFar AND ( i < MaxKey ) DO
          OkSoFar := OkSoFar AND ( key[ i ] = lock[ i ] );
          INC( i );
          END (* WHILE *);
        RETURN OkSoFar;
        END Match;

    VAR
      KeyString     : ShortString;
      SearchPointer : EnvPointer;
      TargetLength  : CARDINAL;

    BEGIN  (* FindString *)
      Concat( key, '=', KeyString );            (* Build '<label>=' field *)
      CapString( KeyString );                (* ..Which is always in caps *)
      SearchPointer := base;          (* Start searching at the beginning *)
      LOOP
        TargetLength := Length( SearchPointer^ );
        IF TargetLength = 0 THEN
          RETURN NIL;                (* Stop on the terminal empty string *)
          END (* IF *);
        IF Match( KeyString, SearchPointer^ ) THEN
          RETURN SearchPointer;
          END (* IF *);
        SearchPointer := ADR( SearchPointer^[ TargetLength + 1 ] );
        END; (* LOOP *)
      END FindString;

  PROCEDURE SearchEnvironment(     key   : ARRAY OF CHAR;
                               VAR value : ARRAY OF CHAR;
                                   level : INTEGER        );
    (* The big one exported to the universe.  See comments in .DEF file *)

    VAR
      ThisPSP   : PSPaddress;
      ThisEnv   : ADDRESS;
      StringLoc : POINTER TO LongString;
      KeySize   : CARDINAL;

    BEGIN (* SearchEnvironment *)
      ThisPSP := SetLevel( level ); (* Grungy DOSism w. SEG-only pointers *)
      ThisEnv.SEGMENT := CARDINAL( ThisPSP^.EnvironmentPointer );
      ThisEnv.OFFSET := 0;
      StringLoc := FindString( key, ThisEnv );
      IF StringLoc = NIL                    (* 'key' currently undefined? *)
        THEN value[ 0 ] := ASCII.nul;    (* If so, return an empty string *)
        ELSE
          KeySize := Length( key ) + 1;
          Copy( StringLoc^, KeySize, Length( StringLoc^ ) - KeySize, value );
        END (* IF *);
      END SearchEnvironment;
(*
   SetEnvironment
*)
  PROCEDURE SetEnvironment(   key : ARRAY OF CHAR;
                            value : ARRAY OF CHAR;
                            level : INTEGER ) : BOOLEAN;
    (* The other big export item.  Specs are in the .DEF file *)

    VAR
      ThisPSP     : PSPaddress;
      ThisEnv     : ADDRESS;
      StringBase  : EnvPointer;
      NewEntry    : LongString;
      OldLength,
      NewLength,
      t1, t2, t3,
      SizeChange  : INTEGER;
      count       : CARDINAL;
      OldPointer,
      NextPointer : POINTER TO EnvironmentString;

    BEGIN (* SetEnvironment *)
      ThisPSP := SetLevel( level );
      ThisEnv.SEGMENT := CARDINAL( ThisPSP^.EnvironmentPointer );
      ThisEnv.OFFSET := 0;
      StringBase := ThisEnv;
      
      IF Length( value ) > 0               (* Are we setting or clearing? *)
        THEN
          Concat( key, '=', NewEntry );    (* Setting -> build the string *)
          CapString( NewEntry );
          Concat( NewEntry, value, NewEntry );
        ELSE NewEntry[ 0 ] := ASCII.nul; (* Clearing -> use a null string *)
        END (* IF *);
      NewLength := Length( NewEntry );
      OldPointer := FindString( key, StringBase );
      IF OldPointer # NIL             (* Is the variable already defined? *)
        THEN OldLength := Length( OldPointer^ );
        ELSE OldLength := 0;
        END;
      t1 := 1 + EndOf( StringBase^ );  
      t2 := ( NewLength - OldLength );
      t3 := SegSize( ThisEnv );
      SizeChange := t1 + t2 - t3;  (* Required change to environment size *)
      IF Resize( StringBase^, SizeChange, ThisPSP )
        THEN RETURN TRUE;                         (* No more room to grow *)
        ELSE
          IF OldLength > 0 THEN                      (* Delete old string *)
            NextPointer := ADR( OldPointer^[ OldLength + 1 ] );
            count := EndOf( NextPointer^ );     (* Gross, but unavoidable *)
            BlockMoveForward( OldPointer, NextPointer, count + 1 );
            END (* IF OldLength *);
          NextPointer := ADR( StringBase^[ EndOf( StringBase^ ) ] );
          Assign( NewEntry, NextPointer^ );     (* Gross, but unavoidable *)
          NextPointer^[ NewLength + 1 ] := ASCII.nul;
          RETURN FALSE;
        END (* IF OK *);
    END SetEnvironment;
(**)
  VAR
    p1, p2 : PSPaddress;

  BEGIN (* Environment *)
    LastLevel := 0;           (* Start with cacheing the local definition *)
    LastPSP := PSPaddress( PSPPtr );
    p1 := LastPSP;                       (* Now determine the chain depth *)
    Depth := 1;                 (* Increment once for first child process *)
    REPEAT
      INC( Depth );
      p2 := p1;
      p1 := Dive( p1 );
      UNTIL p2 = p1;                      (* Stop at self-referential PSP *)
    END Environment.
