MODULE Post;
(************************************************************************)
(*                                                                      *)
(*      Post    operates similarly to the DOS command-line              *)
(*              function 'SET', except that it accepts new              *)
(*              environment strings from STANDARD INPUT                 *)
(*              in the form 'label=value', without the 'SET '           *)
(*              prefix required by COMMAND.COM.  Changed                *)
(*              strings are output in their ORIGINAL forms              *)
(*              to STANDARD OUTPUT, providing a mechanism for           *)
(*              restoring any changes made.                             *)
(*                                                                      *)
(*      Written 20 December 1988        D. C. Sessions                  *)
(*                                                                      *)
(************************************************************************)

  FROM Strings
    IMPORT Length, Copy;
  FROM NumberConversion
    IMPORT StringToInt;
  FROM Environment
    IMPORT SetEnvironment, SearchEnvironment;
  FROM CommandLine
    IMPORT ArgC, ArgV;
  FROM StringOut
    IMPORT WriteString, NewLine;
  FROM StdIn
    IMPORT Read, EOF;
  IMPORT ASCII;

  PROCEDURE ReadLine( VAR line : ARRAY OF CHAR ) : BOOLEAN;
    (* Reads characters until EOF, EOLN, or buffer full *)
    (* Returns TRUE iff EOF *)

    VAR
      index : CARDINAL;
      ch    : CHAR;

    BEGIN (* ReadLine *)
      REPEAT
        Read( ch );
        UNTIL ( ch # ASCII.lf );
      index := 0;
      WHILE ( index < HIGH( line ) ) AND ( ch # ASCII.lf ) AND
            ( ch # ASCII.cr ) AND ( ch # EOF ) DO
         line[ index ] := ch;
         INC( index );
         Read( ch );
         END (* WHILE *);
      line[ index ] := ASCII.nul;
      RETURN ch = EOF;
      END ReadLine;

  PROCEDURE ParseSet(     command : ARRAY OF CHAR;
                      VAR label   : ARRAY OF CHAR;
                      VAR value   : ARRAY OF CHAR );
    (* 'label' is the part of 'command' BEFORE the first '=' *)
    (* 'value' is the part of 'command' AFTER  the first '=' *)

    VAR
      index     : CARDINAL;
      CmdLength : CARDINAL;

    BEGIN (* ParseSet *)
      CmdLength := Length( command );
      index := 0;
      WHILE ( index <= CmdLength ) AND                (* Look for the '=' *)
            ( command[ index ] # '=' ) DO
        INC( index );
        END (* WHILE *);
      Copy( command, 0, index, label );    (* Extract the '<label>=' part *)
      IF index < CmdLength THEN  (* If anything is left, it's the 'value' *)
        INC( index );
        Copy( command, index, CmdLength - index, value );
        END (* IF *);
      END ParseSet;

  TYPE
    ShortString = ARRAY [ 0..127 ] OF CHAR;
    LongString  = ARRAY [ 0..1023 ] OF CHAR;

  VAR
    level    : INTEGER;
    ok,                                   (* Usable command-line <level>? *)
    done     : BOOLEAN;                   (* Command processing complete? *)
    Lstring  : ShortString;                         (* The 'level' option *)
    label    : ShortString;                          (* The variable name *)
    command,                          (* The input buffer for definitions *)
    OldValue,                           (* The previously defined <value> *)
    NewValue : LongString;                  (* The new <value> definition *)

  BEGIN (* Post *)
    ok := FALSE;
    IF ArgC > 0 THEN
      ArgV( 1, Lstring );
      StringToInt( Lstring, level, ok );
      END (* IF *);
    IF NOT ok THEN
      level := -1; (* Default to parent environment *)
      END (* IF *);
    REPEAT
      done := ReadLine( command );              (* Input stream exhausted *)
      IF 1 <= Length( command ) THEN                   (* Minimum is 'a=' *)
        ParseSet( command, label, NewValue );
        SearchEnvironment( label, OldValue, level );
        WriteString( label );                   (* Pass on old definition *)
        WriteString( '=' );
        WriteString( OldValue );
        NewLine;
        done := done OR
                SetEnvironment( label, NewValue, level ); (* No more room *)
        END (* IF *);
      UNTIL done;
    END Post.
